# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import abc
import decimal
import enum
import fractions
import math
import warnings

from . import errors, operators, types

if t.TYPE_CHECKING:
    from . import context, distributions


class Expression(abc.ABC):
    """
    Abstract base class for expressions.
    """

    @abc.abstractmethod
    def infer_type(self, scope: context.Scope) -> types.Type:
        raise NotImplementedError()

    def infer_target_type(self, scope: context.Scope) -> types.Type:
        raise errors.InvalidTypeError(
            f"expression {self} is not a valid assignment target"
        )

    @property
    @abc.abstractmethod
    def children(self) -> t.Sequence[Expression]:
        """
        The direct children of the expression.
        """
        raise NotImplementedError()

    def traverse(self) -> t.Iterator[Expression]:
        """
        Returns an iterator over the subexpressions.
        """
        yield self
        for child in self.children:
            yield from child.traverse()

    @property
    def subexpressions(self) -> t.AbstractSet[Expression]:
        """
        A set of subexpressions.
        """
        return frozenset(self.traverse())

    @property
    def is_sampling_free(self) -> bool:
        """
        Returns whether the expression is *sampling-free*.
        """
        return all(not isinstance(e, Sample) for e in self.traverse())

    @property
    def used_names(self) -> t.AbstractSet[Name]:
        """
        Returns a set of *name expressions* occuring in the expression.
        """
        return frozenset(child for child in self.traverse() if isinstance(child, Name))


class _Leaf(Expression):
    @property
    def children(self) -> t.Sequence[Expression]:
        return ()


class Constant(_Leaf, abc.ABC):
    pass


@d.dataclass(frozen=True)
class BooleanConstant(Constant):
    """
    A boolean constant.

    Attributes
    ----------
    boolean:
        The value of the boolean constant.
    """

    boolean: bool

    def infer_type(self, scope: context.Scope) -> types.Type:
        return types.BOOL


class NumericConstant(Constant, abc.ABC):
    """
    Abstract base class for numeric constants.
    """

    @property
    @abc.abstractmethod
    def as_float(self) -> float:
        """
        The numeric constant as a floating-point number.

        Note that this may result in a loss of precision.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def as_fraction(self) -> fractions.Fraction:
        """
        The numeric constant as a fraction.

        Note that this may result in a loss of precision
        """
        raise NotImplementedError()


TRUE: Expression = BooleanConstant(True)
FALSE: Expression = BooleanConstant(False)


@d.dataclass(frozen=True)
class IntegerConstant(NumericConstant):
    """
    An integer constant.

    Attributes
    ----------
    integer:
        The value of the integer constant.
    """

    integer: int

    def infer_type(self, scope: context.Scope) -> types.Type:
        return types.INT

    @property
    def as_float(self) -> float:
        warnings.warn(
            "converting an integer constant to a float may result in a loss of precision"
        )
        return float(self.integer)

    @property
    def as_fraction(self) -> fractions.Fraction:
        return fractions.Fraction(self.integer)


_NAMED_REAL_MAP: t.Dict[str, NamedReal] = {}


class NamedReal(enum.Enum):
    """
    An enum of named reals.

    Attributes
    ----------
    symbol:
        The mathematical symbol of the real.
    float_value:
        A floating-point approximation of the real.
    """

    PI = "π", math.pi
    """ The number π. """

    E = "e", math.e
    """ The number e. """

    symbol: str
    float_value: float

    def __init__(self, symbol: str, float_value: float) -> None:
        self.symbol = symbol
        self.float_value = float_value
        _NAMED_REAL_MAP[symbol] = self


Real = t.Union[NamedReal, fractions.Fraction]


@d.dataclass(frozen=True)
class RealConstant(NumericConstant):
    """
    A real constant.

    Attributes
    ----------
    real:
        The real (either :class:`NamedReal` or a fraction).
    """

    real: Real

    def infer_type(self, scope: context.Scope) -> types.Type:
        return types.REAL

    @property
    def as_float(self) -> float:
        warnings.warn(
            "converting a real constant to a float may result in a loss of precision"
        )
        if isinstance(self.real, NamedReal):
            return self.real.float_value
        return float(self.real)

    @property
    def as_fraction(self) -> fractions.Fraction:
        if isinstance(self.real, fractions.Fraction):
            return self.real
        else:
            warnings.warn(
                "converting a named real constant do a fraction does result in a loss of precision"
            )
            return fractions.Fraction(self.real.float_value)


@d.dataclass(frozen=True)
class Name(_Leaf):
    """
    A name expression.

    Attributes
    ----------
    identifier:
        The identifier.
    """

    identifier: str

    def infer_type(self, scope: context.Scope) -> types.Type:
        return scope.lookup(self.identifier).typ

    def infer_target_type(self, scope: context.Scope) -> types.Type:
        return self.infer_type(scope)


# XXX: this class should be abstract, however, then it would not type-check
# https://github.com/python/mypy/issues/5374
@d.dataclass(frozen=True)
class BinaryExpression(Expression):
    """
    Abstract base class for binary expressions.

    Attributes
    ----------
    operator:
        The binary operator (:class:`~momba.model.operators.BinaryOperator`).
    left:
        The left operand.
    right:
        The right operand.
    """

    operator: operators.BinaryOperator
    left: Expression
    right: Expression

    @property
    def children(self) -> t.Sequence[Expression]:
        return self.left, self.right

    # XXX: this method shall be implemented by all subclasses
    def infer_type(self, scope: context.Scope) -> types.Type:
        raise NotImplementedError()


class Boolean(BinaryExpression):
    """
    A boolean binary expression.

    Attributes
    ----------
    operator:
        The boolean operator (:class:`~momba.model.operators.BooleanOperator`).
    """

    operator: operators.BooleanOperator

    def infer_type(self, scope: context.Scope) -> types.Type:
        left_type = scope.get_type(self.left)
        if left_type != types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {left_type}")
        right_type = scope.get_type(self.right)
        if right_type != types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {right_type}")
        return types.BOOL


_REAL_RESULT_OPERATORS = {
    operators.ArithmeticBinaryOperator.REAL_DIV,
    operators.ArithmeticBinaryOperator.LOG,
    operators.ArithmeticBinaryOperator.POW,
}


class ArithmeticBinary(BinaryExpression):
    """
    An arithmetic binary expression.

    Attributes
    ----------
    operator:
        The arithmetic operator (:class:`~momba.model.operators.ArithmeticBinaryOperator`).
    """

    operator: operators.ArithmeticBinaryOperator

    def infer_type(self, scope: context.Scope) -> types.Type:
        left_type = scope.get_type(self.left)
        right_type = scope.get_type(self.right)
        if not left_type.is_numeric or not right_type.is_numeric:
            raise errors.InvalidTypeError(
                "operands of arithmetic expressions must have a numeric type"
            )
        is_int = (
            types.INT.is_assignable_from(left_type)
            and types.INT.is_assignable_from(right_type)
            and self.operator not in _REAL_RESULT_OPERATORS
        )
        if is_int:
            return types.INT
        else:
            return types.REAL


class Equality(BinaryExpression):
    """
    An equality binary expression.

    Attributes
    ----------
    operator:
        The equality operator (:class:`~momba.model.operators.EqualityOperator`).
    """

    operator: operators.EqualityOperator

    def get_common_type(self, scope: context.Scope) -> types.Type:
        left_type = scope.get_type(self.left)
        right_type = scope.get_type(self.right)
        if left_type.is_assignable_from(right_type):
            return left_type
        elif right_type.is_assignable_from(left_type):
            return right_type
        raise AssertionError(
            "type-inference should ensure that some of the above is true"
        )

    def infer_type(self, scope: context.Scope) -> types.Type:
        left_type = scope.get_type(self.left)
        right_type = scope.get_type(self.right)
        left_assignable_right = left_type.is_assignable_from(right_type)
        right_assignable_left = right_type.is_assignable_from(left_type)
        if left_assignable_right or right_assignable_left:
            return types.BOOL
        # XXX: JANI specifies that “left and right must be assignable to some common type”
        # not sure whether this implementation reflects this specification
        raise errors.InvalidTypeError(
            "invalid combination of type for equality comparison"
        )


class Comparison(BinaryExpression):
    """
    A comparison expression.

    Attributes
    ----------
    operator:
        The comparison operator (:class:`~momba.model.operators.ComparisonOperator`).
    """

    operator: operators.ComparisonOperator

    def infer_type(self, scope: context.Scope) -> types.Type:
        left_type = scope.get_type(self.left)
        if not left_type.is_numeric:
            raise errors.InvalidTypeError(f"expected numeric type but got {left_type}")
        right_type = scope.get_type(self.right)
        if not right_type.is_numeric:
            raise errors.InvalidTypeError(f"expected numeric type but got {right_type}")
        return types.BOOL


@d.dataclass(frozen=True)
class Conditional(Expression):
    """
    A ternary conditional expression.

    Attributes
    ----------
    condition:
        The condition.
    consequence:
        The consequence to be evaluated if the condition is true.
    alternative:
        The alternative to be evaluated if the condition is false.
    """

    condition: Expression
    consequence: Expression
    alternative: Expression

    @property
    def children(self) -> t.Sequence[Expression]:
        return self.condition, self.consequence, self.alternative

    def infer_type(self, scope: context.Scope) -> types.Type:
        condition_type = scope.get_type(self.condition)
        if condition_type != types.BOOL:
            raise errors.InvalidTypeError(
                f"expected `types.BOOL` but got `{condition_type}`"
            )
        consequence_type = scope.get_type(self.consequence)
        alternative_type = scope.get_type(self.alternative)
        if consequence_type.is_assignable_from(alternative_type):
            return consequence_type
        elif alternative_type.is_assignable_from(consequence_type):
            return alternative_type
        else:
            raise errors.InvalidTypeError(
                "invalid combination of consequence and alternative types"
            )


# XXX: this class should be abstract, however, then it would not type-check
# https://github.com/python/mypy/issues/5374
@d.dataclass(frozen=True)
class UnaryExpression(Expression, abc.ABC):
    """
    Base class of all unary expressions.

    Attributes
    ----------
    operator:
        The unary operator (:class:`~momba.model.operators.UnaryOperator`).
    operand:
        The operand.
    """

    operator: operators.UnaryOperator
    operand: Expression

    @property
    def children(self) -> t.Sequence[Expression]:
        return (self.operand,)

    # XXX: this method shall be implemented by all subclasses
    def infer_type(self, scope: context.Scope) -> types.Type:
        raise NotImplementedError()


class ArithmeticUnary(UnaryExpression):
    """
    An arithmetic unary expression.

    Attributes
    ----------
    operator:
        The arithmetic operator (:class:`~momba.model.operators.ArithmeticUnaryOperator`).
    """

    operator: operators.ArithmeticUnaryOperator

    def infer_type(self, scope: context.Scope) -> types.Type:
        operand_type = scope.get_type(self.operand)
        if not operand_type.is_numeric:
            raise errors.InvalidTypeError(
                f"expected a numeric type but got {operand_type}"
            )
        return self.operator.infer_result_type(operand_type)


class Not(UnaryExpression):
    """
    Logical negation.

    Attributes
    ----------
    operator:
        The logical negation operator (:class:`~momba.model.operators.NotOperator`).
    """

    operator: operators.NotOperator

    def infer_type(self, scope: context.Scope) -> types.Type:
        operand_type = scope.get_type(self.operand)
        if operand_type != types.BOOL:
            raise errors.InvalidTypeError(
                f"expected `types.BOOL` but got {operand_type}"
            )
        return types.BOOL


@d.dataclass(frozen=True)
class Sample(Expression):
    """
    A sample expression.

    Attributes
    ----------
    distribution:
        The type of the distribution to sample from
        (:class:`~momba.model.distributions.DistributionType`).
    arguments:
        The arguments to the distribution.
    """

    distribution: distributions.DistributionType
    arguments: t.Sequence[Expression]

    def __post_init__(self) -> None:
        if len(self.arguments) != self.distribution.arity:
            raise errors.InvalidTypeError(
                f"distribution {self.distribution} requires {self.distribution.arity} "
                f"arguments but {len(self.arguments)} were given"
            )

    @property
    def children(self) -> t.Sequence[Expression]:
        return self.arguments

    def infer_type(self, scope: context.Scope) -> types.Type:
        # we already know that the arity of the parameters and arguments match
        for argument, parameter_type in zip(
            self.arguments, self.distribution.parameter_types
        ):
            argument_type = scope.get_type(argument)
            if not parameter_type.is_assignable_from(argument_type):
                raise errors.InvalidTypeError(
                    f"parameter type `{parameter_type}` is not assignable "
                    f"from argument type `{argument_type}`"
                )
        return self.distribution.result_type


# requires JANI extension `nondet-selection`
@d.dataclass(frozen=True)
class Selection(Expression):
    """
    A non-deterministic selection expression.

    Attributes
    ----------
    variable:
        The identifier to select over.
    condition:
        The condition that should be satisfied.
    """

    variable: str
    condition: Expression

    def infer_type(self, scope: context.Scope) -> types.Type:
        condition_scope = scope.create_child_scope()
        condition_scope.declare_variable(self.variable, typ=types.REAL)
        condition_type = condition_scope.get_type(self.condition)
        if condition_type != types.BOOL:
            raise errors.InvalidTypeError("condition must have type `types.BOOL`")
        return types.REAL

    @property
    def children(self) -> t.Sequence[Expression]:
        return (self.condition,)


@d.dataclass(frozen=True)
class Derivative(Expression):
    """
    Derivative of a continuous variable.

    Attributes
    ==========
    identifier:
        The continuous variable.
    """

    identifier: str

    def infer_type(self, scope: context.Scope) -> types.Type:
        return types.REAL

    @property
    def children(self) -> t.Sequence[Expression]:
        return ()


@d.dataclass(frozen=True)
class ArrayAccess(Expression):
    """
    An array access expression.

    Attributes
    ----------
    array:
        The array to access.
    index:
        The index where to access the array.
    """

    array: Expression
    index: Expression

    @property
    def children(self) -> t.Sequence[Expression]:
        return self.array, self.index

    def infer_type(self, scope: context.Scope) -> types.Type:
        array_type = scope.get_type(self.array)
        # TODO: check the type of the index
        if isinstance(array_type, types.ArrayType):
            return array_type.element
        else:
            raise errors.InvalidTypeError("array of array access must have array type")

    def infer_target_type(self, scope: context.Scope) -> types.Type:
        return self.infer_type(scope)


@d.dataclass(frozen=True)
class ArrayValue(Expression):
    """
    An array value expression.

    Attributes
    ----------
    elements:
        The elements of the array to construct.
    """

    elements: t.Tuple[Expression, ...]

    def __post_init__(self) -> None:
        if not self.elements:
            raise errors.ModelingError("array value expression needs to have elements")

    @property
    def children(self) -> t.Sequence[Expression]:
        return self.elements

    def infer_type(self, scope: context.Scope) -> types.Type:
        common_type: t.Optional[types.Type] = None
        for element in self.elements:
            element_type = scope.get_type(element)
            if common_type is None:
                common_type = element_type
            elif element_type.is_assignable_from(common_type):
                common_type = element_type
            elif not common_type.is_assignable_from(element_type):
                raise errors.InvalidTypeError(
                    "element types are not assignable to a common type"
                )
        assert common_type is not None
        return types.array_of(common_type)


@d.dataclass(frozen=True)
class ArrayConstructor(Expression):
    """
    An array constructor expression.

    Attributes
    ----------
    variable:
        The identifier to range over.
    length:
        The length of the array.
    expression:
        The expression to compute the elements of the array.
    """

    variable: str
    length: Expression
    expression: Expression

    @property
    def children(self) -> t.Sequence[Expression]:
        return self.length, self.expression

    def _create_scope(self, scope: context.Scope) -> context.Scope:
        child_scope = scope.create_child_scope()
        child_scope.declare_constant(self.variable, types.INT)
        return child_scope

    def infer_type(self, scope: context.Scope) -> types.Type:
        if not types.INT.is_assignable_from(scope.get_type(self.length)):
            raise errors.InvalidTypeError(
                "length of array constructor has to be an integer"
            )
        return types.array_of(self._create_scope(scope).get_type(self.expression))


class Trigonometric(UnaryExpression):
    """
    A trigonometric expression.

    Attributes
    ----------
    operator:
        The trigonometric function to apply
        (:class:`~momba.model.operators.TrigonometricFunction`).
    """

    operator: operators.TrigonometricFunction

    def infer_type(self, scope: context.Scope) -> types.Type:
        operand_type = scope.get_type(self.operand)
        if not operand_type.is_numeric:
            raise errors.InvalidTypeError(
                "expected numeric type for operand of trigonometric function"
            )
        return types.REAL


RealValue = t.Union[t.Literal["π", "e"], float, fractions.Fraction, decimal.Decimal]
NumericValue = t.Union[int, RealValue]
Value = t.Union[bool, NumericValue]

ValueOrExpression = t.Union[Expression, Value]


class ConversionError(ValueError):
    """
    Unable to convert the provided value.
    """


def ensure_expr(value_or_expression: ValueOrExpression) -> Expression:
    """
    Takes a Python value or expression and returns an expression.

    Implicitly converts floats, integers, and booleans to constant expressions.

    Raises :class:`~momba.model.expressions.ConversionError` if the conversion fails.
    """
    if isinstance(value_or_expression, Expression):
        return value_or_expression
    elif isinstance(value_or_expression, bool):
        return BooleanConstant(value_or_expression)
    elif isinstance(value_or_expression, int):
        return IntegerConstant(value_or_expression)
    elif isinstance(value_or_expression, (float, fractions.Fraction, decimal.Decimal)):
        return RealConstant(fractions.Fraction(value_or_expression))
    elif isinstance(value_or_expression, str):
        try:
            return RealConstant(_NAMED_REAL_MAP[value_or_expression])
        except KeyError:
            pass
    raise ConversionError(
        f"unable to convert Python value {value_or_expression!r} to expression"
    )


def ite(
    condition: ValueOrExpression,
    consequence: ValueOrExpression,
    alternative: ValueOrExpression,
) -> Expression:
    """
    Constructs a conditional expression implicitly converting the arguments.
    """
    return Conditional(
        ensure_expr(condition), ensure_expr(consequence), ensure_expr(alternative)
    )


BinaryConstructor = t.Callable[[ValueOrExpression, ValueOrExpression], Expression]


def _boolean_binary_expression(
    operator: operators.BooleanOperator, expressions: t.Sequence[ValueOrExpression]
) -> Expression:
    if len(expressions) == 1:
        return ensure_expr(expressions[0])
    result = Boolean(
        operator,
        ensure_expr(expressions[0]),
        ensure_expr(expressions[1]),
    )
    for operand in expressions[2:]:
        result = Boolean(operator, result, ensure_expr(operand))
    return result


def logic_not(operand: ValueOrExpression) -> Expression:
    """
    Constructs a logical negation expression.
    """
    return Not(operators.NotOperator.NOT, ensure_expr(operand))


def logic_any(*expressions: ValueOrExpression) -> Expression:
    """
    Constructs a disjunction over the provided expressions.

    Returns :attr:`~momba.model.expressions.FALSE` if there are no expressions.
    """
    if len(expressions) == 0:
        return BooleanConstant(False)
    return logic_or(*expressions)


def logic_or(*expressions: ValueOrExpression) -> Expression:
    """
    Constructs a disjunction over the provided expressions.
    """
    return _boolean_binary_expression(operators.BooleanOperator.OR, expressions)


def logic_all(*expressions: ValueOrExpression) -> Expression:
    """
    Constructs a conjunction over the provided expressions.

    Returns :attr:`~momba.model.expressions.TRUE` if there are no expressions.
    """
    if len(expressions) == 0:
        return BooleanConstant(True)
    return logic_and(*expressions)


def logic_and(*expressions: ValueOrExpression) -> Expression:
    """
    Constructs a conjunction over the provided expressions.
    """
    return _boolean_binary_expression(operators.BooleanOperator.AND, expressions)


def logic_xor(*expressions: ValueOrExpression) -> Expression:
    """
    Constructs an exclusive disjunction over the provided expressions.
    """
    return _boolean_binary_expression(operators.BooleanOperator.XOR, expressions)


def logic_implies(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs a logical implication.
    """
    return Boolean(
        operators.BooleanOperator.IMPLY, ensure_expr(left), ensure_expr(right)
    )


def logic_equiv(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs a logical equivalence.
    """
    return Boolean(
        operators.BooleanOperator.EQUIV, ensure_expr(left), ensure_expr(right)
    )


def equals(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs an *equality* expression.
    """
    return Equality(
        operators.EqualityOperator.EQ, ensure_expr(left), ensure_expr(right)
    )


def not_equals(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs an *inequality* expression.
    """
    return Equality(
        operators.EqualityOperator.NEQ, ensure_expr(left), ensure_expr(right)
    )


def less(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs a *less than* expression.
    """
    return Comparison(
        operators.ComparisonOperator.LT, ensure_expr(left), ensure_expr(right)
    )


def less_or_equal(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs a *less than or equal to* expression.
    """
    return Comparison(
        operators.ComparisonOperator.LE, ensure_expr(left), ensure_expr(right)
    )


def greater(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs a *greater than* expression.
    """
    return Comparison(
        operators.ComparisonOperator.GT, ensure_expr(left), ensure_expr(right)
    )


def greater_or_equal(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs a *greater than or equal to* expression.
    """
    return Comparison(
        operators.ComparisonOperator.GE, ensure_expr(left), ensure_expr(right)
    )


def add(left: ValueOrExpression, right: ValueOrExpression) -> Expression:
    """
    Constructs an arithmetic addition expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.ADD, ensure_expr(left), ensure_expr(right)
    )


def sub(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs an arithmetic substraction expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.SUB, ensure_expr(left), ensure_expr(right)
    )


def mul(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs an arithmetic multiplication expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.MUL, ensure_expr(left), ensure_expr(right)
    )


def mod(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs an euclidean remainder expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.MOD, ensure_expr(left), ensure_expr(right)
    )


def real_div(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs a real-division expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.REAL_DIV,
        ensure_expr(left),
        ensure_expr(right),
    )


def log(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs a logarithm expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.LOG, ensure_expr(left), ensure_expr(right)
    )


def power(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs a power expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.POW, ensure_expr(left), ensure_expr(right)
    )


def minimum(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs a minimum expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.MIN, ensure_expr(left), ensure_expr(right)
    )


def maximum(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs a maximum expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.MAX, ensure_expr(left), ensure_expr(right)
    )


def floor_div(left: ValueOrExpression, right: ValueOrExpression) -> BinaryExpression:
    """
    Constructs an euclidean division expression.
    """
    return ArithmeticBinary(
        operators.ArithmeticBinaryOperator.FLOOR_DIV,
        ensure_expr(left),
        ensure_expr(right),
    )


UnaryConstructor = t.Callable[[ValueOrExpression], Expression]


def floor(operand: ValueOrExpression) -> Expression:
    """
    Constructs a floor expression.
    """
    return ArithmeticUnary(
        operators.ArithmeticUnaryOperator.FLOOR, ensure_expr(operand)
    )


def ceil(operand: ValueOrExpression) -> Expression:
    """
    Constructs a ceil expression.
    """
    return ArithmeticUnary(operators.ArithmeticUnaryOperator.CEIL, ensure_expr(operand))


def absolute(operand: ValueOrExpression) -> Expression:
    """
    Constructs an absolute value expression.
    """
    return ArithmeticUnary(operators.ArithmeticUnaryOperator.ABS, ensure_expr(operand))


def sgn(operand: ValueOrExpression) -> Expression:
    """
    Constructs a sign expression.
    """
    return ArithmeticUnary(operators.ArithmeticUnaryOperator.SGN, ensure_expr(operand))


def trunc(operand: ValueOrExpression) -> Expression:
    """
    Constructs a truncate expression.
    """
    return ArithmeticUnary(operators.ArithmeticUnaryOperator.TRC, ensure_expr(operand))


def name(identifier: str) -> Name:
    """
    Constructs a name expression.
    """
    return Name(identifier)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import abc
import fractions

from .. import model


Properties = t.Mapping[str, model.Expression]
Result = t.Mapping[str, fractions.Fraction]


class Checker(abc.ABC):
    """An abstract class for model checkers."""

    @property
    def description(self) -> str:
        return self.__class__.__name__

    @abc.abstractmethod
    def check(
        self,
        network: model.Network,
        *,
        properties: t.Optional[Properties] = None,
        property_names: t.Optional[t.Iterable[str]] = None,
    ) -> Result:
        """Model checks the given properties on the network."""
        raise NotImplementedError()


class CrossCheckError(Exception):
    pass


class PropertyMismatchError(Exception):
    pass


class DeltaExceededError(Exception):
    pass


@d.dataclass(frozen=True, eq=False)
class CrossChecker(Checker):
    checkers: t.Sequence[Checker]

    allowed_delta: fractions.Fraction = fractions.Fraction("1e-4")

    @property
    def description(self) -> str:
        return f"CrossChecker (Δ = {float(self.allowed_delta)})"

    def check(
        self,
        network: model.Network,
        *,
        properties: t.Optional[Properties] = None,
        property_names: t.Optional[t.Iterable[str]] = None,
    ) -> Result:
        assert self.checkers, "checkers for cross-checking must be non-empty"
        results = [
            checker.check(network, properties=properties, property_names=property_names)
            for checker in self.checkers
        ]
        names = frozenset(results[0].keys())
        for result in results:
            if names != result.keys():
                raise PropertyMismatchError(
                    "checkers did not provide results for the same properties"
                )
        for first in range(len(self.checkers)):
            for second in range(first + 1, len(self.checkers)):
                for name in names:
                    first_value = results[first][name]
                    second_value = results[second][name]
                    if abs(first_value - second_value) > self.allowed_delta:
                        raise DeltaExceededError(
                            f"allowed delta of {self.allowed_delta} has been exceeded for {name}"
                        )
        return results[0]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import abc

from .. import model

from ._engine import engine as _engine

from . import zones

from .translator import Translation, translate_network

if t.TYPE_CHECKING:
    from .explore import Parameters


class InvalidModelType(Exception):
    pass


T = t.TypeVar("T")


@d.dataclass(frozen=True)
class CompiledNetwork:
    translation: Translation
    internal: t.Any


class TimeType(abc.ABC):
    """
    Base class for time representations.
    """

    @staticmethod
    @abc.abstractmethod
    def compile(
        network: model.Network, *, parameters: Parameters = None
    ) -> CompiledNetwork:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def load_valuations(cls: t.Type[T], valuations: t.Any) -> T:
        raise NotImplementedError()


class DiscreteTime(TimeType):
    """
    A representation of time without continuous-time clocks.
    """

    @staticmethod
    def compile(
        network: model.Network, *, parameters: Parameters = None
    ) -> CompiledNetwork:
        translation = translate_network(network, parameters=parameters)
        if not network.ctx.model_type.is_untimed:
            raise InvalidModelType(
                f"{network.ctx.model_type} is not a discrete time model type"
            )
        return CompiledNetwork(
            translation, _engine.Explorer.new_no_clocks(translation.json_network)
        )

    @classmethod
    def load_valuations(cls, valuations: t.Any) -> DiscreteTime:
        return cls()


@d.dataclass(frozen=True)
class GlobalTime(TimeType):
    zone: zones.Zone[float]

    @staticmethod
    def compile(
        network: model.Network, *, parameters: Parameters = None
    ) -> CompiledNetwork:
        translation = translate_network(
            network, parameters=parameters, global_clock=True
        )
        return CompiledNetwork(
            translation, _engine.Explorer.new_global_time(translation.json_network)
        )

    @classmethod
    def load_valuations(cls, valuations: t.Any) -> GlobalTime:
        return cls(zones._wrap_zone(valuations, zones.ZoneF64))


@d.dataclass(frozen=True)
class ZoneF64(TimeType):
    zone: zones.Zone[float]

    @staticmethod
    def compile(
        network: model.Network, *, parameters: Parameters = None
    ) -> CompiledNetwork:
        translation = translate_network(
            network, parameters=parameters, global_clock=False
        )
        return CompiledNetwork(
            translation, _engine.Explorer.new_global_time(translation.json_network)
        )

    @classmethod
    def load_valuations(cls, valuations: t.Any) -> ZoneF64:
        return cls(zones._wrap_zone(valuations, zones.ZoneF64))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

"""
An implementation of *Difference Bound Matrices* (DBMs) in pure Python.
"""

from __future__ import annotations

import dataclasses as d
import typing as t

import abc
import fractions
import itertools
import math

from .intervals import Interval


NumberType = t.Union[int, float, fractions.Fraction]


class InvalidBoundError(ValueError):
    pass


@d.dataclass(frozen=True, order=False)
class Bound:
    constant: NumberType
    is_strict: bool = False

    def __post_init__(self) -> None:
        if self.is_infinity and not self.is_strict:
            raise InvalidBoundError("bound with constant ∞ must be strict")

    @classmethod
    def less_than(cls, constant: NumberType) -> Bound:
        return Bound(constant, is_strict=True)

    @classmethod
    def less_or_equal(cls, constant: NumberType) -> Bound:
        return Bound(constant, is_strict=False)

    @property
    def is_infinity(self) -> bool:
        return math.isinf(self.constant)

    @property
    def is_integer(self) -> bool:
        if isinstance(self.constant, int):
            return True
        elif isinstance(self.constant, float):
            return self.constant.is_integer()
        else:
            assert isinstance(self.constant, fractions.Fraction)
            return self.constant.denominator == 1

    def add(self, other: Bound) -> Bound:
        if self.is_strict or other.is_strict:
            return Bound(self.constant + other.constant, is_strict=True)
        return Bound(self.constant + other.constant, is_strict=False)

    def __str__(self) -> str:
        operator = "<" if self.is_strict else "≤"
        constant = "∞" if self.is_infinity else str(self.constant)
        return f"{operator} {constant}"

    def __lt__(self, other: Bound) -> bool:
        if not isinstance(other, Bound):
            return NotImplemented
        if self.is_strict is other.is_strict:
            return self.constant < other.constant
        elif self.is_strict and not other.is_strict:
            return self.constant <= other.constant
        assert not self.is_strict and other.is_strict
        return other.constant > self.constant

    def __le__(self, other: Bound) -> bool:
        if not isinstance(other, Bound):
            return NotImplemented
        return self == other or self < other


# perform some basic correctness checks
assert not (Bound(3, is_strict=False) < Bound(3, is_strict=False))
assert not (Bound(3, is_strict=False) < Bound(3, is_strict=True))
assert Bound(3, is_strict=True) < Bound(3, is_strict=False)
assert not (Bound(3, is_strict=True) < Bound(3, is_strict=True))
assert Bound(-3, is_strict=True) < Bound(2, is_strict=True)
assert Bound(3, is_strict=True) > Bound(2, is_strict=False)


@d.dataclass(frozen=True)
class ZeroClock:
    def __str__(self) -> str:
        return "0"


@d.dataclass(frozen=True)
class NamedClock:
    name: str

    def __str__(self) -> str:
        return f"Clock({self.name})"


ZERO_CLOCK = ZeroClock()


ClockT = t.TypeVar("ClockT", bound=t.Hashable)


@d.dataclass(frozen=True)
class Difference(t.Generic[ClockT]):
    left: t.Union[ClockT, ZeroClock]
    right: t.Union[ClockT, ZeroClock]

    def __str__(self) -> str:
        return f"{self.left} - {self.right}"

    def bound(self, bound: Bound) -> Constraint[ClockT]:
        return Constraint(self, bound=bound)

    def less_than(self, constant: NumberType) -> Constraint[ClockT]:
        return self.bound(Bound.less_than(constant))

    def less_or_equal(self, constant: NumberType) -> Constraint[ClockT]:
        return self.bound(Bound.less_or_equal(constant))


@d.dataclass(frozen=True)
class Constraint(t.Generic[ClockT]):
    difference: Difference[ClockT]
    bound: Bound

    def __str__(self) -> str:
        return f"{self.difference} {self.bound}"

    @property
    def clocks(self) -> t.AbstractSet[t.Union[ClockT, ZeroClock]]:
        return {self.difference.left, self.difference.right}

    @property
    def left(self) -> t.Union[ClockT, ZeroClock]:
        return self.difference.left

    @property
    def right(self) -> t.Union[ClockT, ZeroClock]:
        return self.difference.right

    @property
    def constant(self) -> NumberType:
        return self.bound.constant

    @property
    def is_strict(self) -> bool:
        return self.bound.is_strict


def difference(
    clock: t.Union[ZeroClock, ClockT], other: t.Union[ZeroClock, ClockT]
) -> Difference[ClockT]:
    return Difference(clock, other)


_INFINITY_BOUND = Bound.less_than(float("inf"))
_ZERO_BOUND = Bound.less_or_equal(0)

_Matrix = t.Dict[Difference[ClockT], Bound]


Clocks = t.AbstractSet[t.Union[ClockT, ZeroClock]]

_FrozenClocks = t.FrozenSet[t.Union[ClockT, ZeroClock]]


def _create_matrix(clocks: Clocks[ClockT]) -> _Matrix[ClockT]:
    matrix: _Matrix[ClockT] = {}
    for clock in clocks:
        # the value of clocks is always positive, and …
        matrix[difference(ZERO_CLOCK, clock)] = _ZERO_BOUND
        # … the difference of each clock and itself is zero
        matrix[difference(clock, clock)] = _ZERO_BOUND
    return matrix


def _freeze_clocks(clocks: Clocks[ClockT]) -> _FrozenClocks[ClockT]:
    return frozenset(clocks).union({ZERO_CLOCK})


class InvalidClockError(ValueError):
    pass


class AbstractDBM(abc.ABC, t.Generic[ClockT]):
    @property
    @abc.abstractmethod
    def constraints(self) -> t.AbstractSet[Constraint[ClockT]]:
        raise NotImplementedError()


@d.dataclass(frozen=True)
class FrozenDBM(AbstractDBM[ClockT], t.Generic[ClockT]):
    _clocks: _FrozenClocks[ClockT]
    _constraints: t.FrozenSet[Constraint[ClockT]]

    def create_dbm(self) -> DBM[ClockT]:
        dbm = DBM.create_unconstrained(self._clocks)
        dbm.constrain(*self._constraints)
        return dbm

    @property
    def constraints(self) -> t.AbstractSet[Constraint[ClockT]]:
        return self._constraints


@d.dataclass
class DBM(AbstractDBM[ClockT], t.Generic[ClockT]):
    _clocks: _FrozenClocks[ClockT]
    _matrix: _Matrix[ClockT]

    @classmethod
    def create_unconstrained(cls, clocks: Clocks[ClockT]) -> DBM[ClockT]:
        """Creates a DBM without any constraints."""
        frozen_clocks = _freeze_clocks(clocks)
        return DBM(frozen_clocks, _create_matrix(frozen_clocks))

    @classmethod
    def create_zero(cls, clocks: Clocks[ClockT]) -> DBM[ClockT]:
        """Creates a DBM where all clocks are constraint to be zero."""
        frozen_clocks = _freeze_clocks(clocks)
        matrix = _create_matrix(frozen_clocks)
        for clock in frozen_clocks:
            matrix[difference(clock, ZERO_CLOCK)] = _ZERO_BOUND
        dbm = DBM(frozen_clocks, matrix)
        dbm._canonicalize()
        return dbm

    def __post_init__(self) -> None:
        assert self._assert_invariant()

    def _assert_invariant(self) -> bool:
        for clock in self._clocks:
            assert self._matrix[difference(ZERO_CLOCK, clock)] <= _ZERO_BOUND
            assert self._matrix[difference(clock, clock)] <= _ZERO_BOUND
        return True

    def _canonicalize(self) -> None:
        for x, y, z in itertools.product(self._clocks, repeat=3):
            xy_bound = self.get_bound(x, y)
            yz_bound = self.get_bound(y, z)
            xz_bound = self.get_bound(x, z)
            if xz_bound > xy_bound.add(yz_bound):
                self._matrix[difference(x, z)] = xy_bound.add(yz_bound)
        assert self._assert_invariant()

    def _set_bound(
        self,
        left: t.Union[ClockT, ZeroClock],
        right: t.Union[ClockT, ZeroClock],
        bound: Bound,
    ) -> None:
        self._matrix[difference(left, right)] = bound
        assert self._assert_invariant()

    def get_bound(
        self, left: t.Union[ClockT, ZeroClock], right: t.Union[ClockT, ZeroClock]
    ) -> Bound:
        return self._matrix.get(difference(left, right), _INFINITY_BOUND)

    def extend_clocks(self, clocks: Clocks[ClockT]) -> DBM[ClockT]:
        matrix = dict(self._matrix)
        for clock in clocks:
            matrix[difference(ZERO_CLOCK, clock)] = _ZERO_BOUND
            matrix[difference(clock, clock)] = _ZERO_BOUND
        return DBM(_freeze_clocks(clocks | self.clocks), matrix)

    @property
    def clocks(self) -> t.AbstractSet[t.Union[ClockT, ZeroClock]]:
        return self._clocks

    @property
    def constraints(self) -> t.AbstractSet[Constraint[ClockT]]:
        return {Constraint(diff, bound) for diff, bound in self._matrix.items()}

    @property
    def is_empty(self) -> bool:
        for clock in self._clocks:
            if self.get_bound(clock, clock) < _ZERO_BOUND:
                return True
        return False

    def freeze(self) -> FrozenDBM[ClockT]:
        return FrozenDBM(self._clocks, frozenset(self.constraints))

    def get_interval(self, clock: ClockT) -> Interval[NumberType]:
        lower_bound = self.get_bound(ZERO_CLOCK, clock)
        upper_bound = self.get_bound(clock, ZERO_CLOCK)
        return Interval(
            -lower_bound.constant,
            upper_bound.constant,
            infimum_included=not lower_bound.is_strict,
            supremum_included=not upper_bound.is_strict,
        )

    def _constrain(self, difference: Difference[ClockT], by: Bound) -> None:
        if difference.left not in self._clocks or difference.right not in self._clocks:
            raise InvalidClockError(
                f"unable to constrain with {difference}: unknown clocks"
            )
        if by < self._matrix.get(difference, _INFINITY_BOUND):
            self._matrix[difference] = by

    def copy(self) -> DBM[ClockT]:
        return DBM(_clocks=self._clocks, _matrix=self._matrix.copy())

    def reset(self, clock: ClockT, value: NumberType = 0) -> None:
        assert not self.is_empty
        upper_bound = Bound.less_or_equal(value)
        lower_bound = Bound.less_or_equal(-value)
        for other in self._clocks:
            if other == clock:
                continue
            self._set_bound(
                clock, other, upper_bound.add(self.get_bound(ZERO_CLOCK, other))
            )
            self._set_bound(
                other, clock, self.get_bound(other, ZERO_CLOCK).add(lower_bound)
            )

    @t.overload
    def constrain(self, *constraints: Constraint[ClockT]) -> None:
        pass

    @t.overload
    def constrain(self, *, difference: Difference[ClockT], by: Bound) -> None:
        pass

    def constrain(
        self,
        *constraints: Constraint[ClockT],
        difference: t.Optional[Difference[ClockT]] = None,
        by: t.Optional[Bound] = None,
    ) -> None:
        try:
            for constraint in constraints:
                self._constrain(constraint.difference, constraint.bound)
            if difference is not None:
                assert by is not None, "use a static type checker!"
                self._constrain(difference, by)
        finally:
            # XXX: not optimal; use the knowledge about which clocks have been touched
            self._canonicalize()

    def advance_upper_bounds(self, time_delta: t.Optional[NumberType] = None) -> None:
        """
        Advances the upper bounds of all clocks by the given amount of time.
        """
        for clock in self._clocks:
            if clock == ZERO_CLOCK:
                continue
            if time_delta is None:
                # delete the upper bound of the clock
                del self._matrix[difference(clock, ZERO_CLOCK)]
            else:
                # advance the upper bound by the given time
                self._set_bound(
                    clock,
                    ZERO_CLOCK,
                    self.get_bound(clock, ZERO_CLOCK).add(
                        Bound.less_or_equal(time_delta)
                    ),
                )
        self._canonicalize()

    def advance_lower_bounds(self, delta: NumberType) -> None:
        """
        Advances the lower bounds of all clocks by the given amount of time.
        """
        for clock in self._clocks:
            if clock == ZERO_CLOCK:
                continue
            self._set_bound(
                ZERO_CLOCK,
                clock,
                self.get_bound(ZERO_CLOCK, clock).add(Bound.less_or_equal(-delta)),
            )
        self._canonicalize()

    def future(self) -> None:
        """
        Removes the upper bounds on all clocks.
        """
        self.advance_upper_bounds()

    def past(self) -> None:
        """
        Sets the lower bound of all clocks to zero.
        """
        try:
            for clock in self._clocks:
                if clock == ZERO_CLOCK:
                    continue
                self._set_bound(ZERO_CLOCK, clock, Bound.less_or_equal(0))
        finally:
            self._canonicalize()

    def intersect(self, other: DBM[ClockT]) -> None:
        assert other._clocks <= self._clocks
        try:
            for difference, bound in other._matrix.items():
                self._constrain(difference, bound)
        finally:
            self._canonicalize()


def print_constraints(dbm: AbstractDBM[ClockT]) -> None:
    for constraint in dbm.constraints:
        print(constraint)


def intersect(left: DBM[ClockT], right: DBM[ClockT]) -> DBM[ClockT]:
    result = left.copy()
    result.intersect(right)
    return result


def create_clock(name: str) -> NamedClock:
    return NamedClock(name)


def create_clocks(*names: str) -> t.Sequence[NamedClock]:
    return list(map(create_clock, names))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import enum
import re


KEYWORDS = {
    "model_type",
    "transient",
    "assign",
    "variable",
    "constant",
    "action",
    "comment",
    "parameter",
    "automaton",
    "network",
    "initial",
    "location",
    "invariant",
    "edge",
    "from",
    "to",
    "guard",
    "rate",
    "probability",
    "restrict",
    "instance",
    "input",
    "enable",
    "composition",
    "synchronize",
    "metadata",
    "true",
    "false",
}

PRIMITIVE_TYPES = {"bool", "int", "real", "clock", "continuous"}


class TokenType(enum.Enum):
    regex: str
    pseudo: bool

    COMMENT = r"#.*|\(\*(.|\s)*?\*\)"

    READ = r"\?>"
    WRITE = r"<!"

    ASSIGN = r":="

    SLASH_SLASH = r"//"

    NAMED_REAL = r"real\[\w+\]"

    TAU = r"τ"

    ARROW = r"→|->"

    LOGIC_AND = r"∧|and"
    LOGIC_OR = r"∨|or"
    LOGIC_XOR = r"⊕|xor"
    LOGIC_IMPLIES = r"⇒|==>|=>"
    LOGIC_EQUIV = r"⇔|<=>"
    LOGIC_NOT = r"¬|not"

    COMP_LE = r"≤|<="
    COMP_GE = r"≥|>="
    COMP_LT = r"<"
    COMP_GT = r">"

    COMP_EQ = r"==|="
    COMP_NEQ = r"≠|!="

    POWER = r"\*\*"

    PLUS = r"\+"
    STAR = r"\*"
    PERCENTAGE = r"%"
    SLASH = r"/"
    QUESTIONMARK = r"\?"
    PIPE = r"\|"

    COMMA = r","
    COLLON = r":"

    LEFT_BRACE = r"\{"
    RIGHT_BRACE = r"\}"

    FILTER_LEFT = r"\(\{"
    FILTER_RIGHT = r"\}\)"

    LEFT_PAR = r"\("
    RIGHT_PAR = r"\)"

    LEFT_BRACKET = r"\["
    RIGHT_BRACKET = r"\]"

    STRING = r"\"(?P<string>[^\"]*)\""

    REAL = r"\d+\.\d+"
    INTEGER = r"\d+"

    MINUS = r"-"

    MACRO = r"\$(?P<macro_name>\w+)"

    IDENTIFIER = r"\w+"

    WHITESPACE = r"\s+"

    ERROR = r"."

    KEYWORD = r"KEYWORD", True

    INDENT = r"INDENT", True
    DEDENT = r"DEDENT", True

    END_OF_FILE = r"END_OF_FILE", True

    def __init__(self, regex: str, pseudo: bool = False) -> None:
        self.regex = regex
        self.pseudo = pseudo


@d.dataclass(frozen=True)
class Token:
    token_type: TokenType

    text: str
    match: t.Match[str]

    start_row: int
    start_column: int

    end_row: int
    end_column: int


_lexer_regex = re.compile(
    r"|".join(
        f"(?P<{token_type.name}>{token_type.value})"
        for token_type in TokenType
        if not token_type.pseudo
    )
)


_OPEN_PARENTHESIS = {TokenType.LEFT_BRACE, TokenType.LEFT_BRACKET, TokenType.LEFT_PAR}
_CLOSE_PARENTHESIS = {
    TokenType.RIGHT_BRACE,
    TokenType.RIGHT_BRACKET,
    TokenType.RIGHT_PAR,
}


class LexerError(Exception):
    pass


def lex(code: str, *, row: int = 0, column: int = 0) -> t.Iterator[Token]:
    indents: t.List[int] = [0]
    parenthesis_level = 0
    end_row = row
    end_column = column
    for match in _lexer_regex.finditer(code):
        assert isinstance(match.lastgroup, str)
        token_type = TokenType[match.lastgroup]
        text = match[0]
        if token_type is TokenType.IDENTIFIER and text in KEYWORDS:
            token_type = TokenType.KEYWORD
        if token_type is TokenType.ERROR:
            raise LexerError(f"unexpected character {text!r} at {row}:{column}")
        if "\n" in text:
            end_row = row + text.count("\n")
            end_column = len(text.rpartition("\n")[2])
            if token_type is TokenType.WHITESPACE and parenthesis_level == 0:
                if end_column > indents[-1]:
                    indents.append(end_column)
                    yield Token(
                        TokenType.INDENT,
                        "",
                        match,
                        end_row,
                        end_column,
                        end_row,
                        end_column,
                    )
                else:
                    while end_column < indents[-1]:
                        indents.pop()
                        yield Token(
                            TokenType.DEDENT,
                            "",
                            match,
                            end_row,
                            end_column,
                            end_row,
                            end_column,
                        )
                    if indents[-1] != end_column:
                        raise LexerError("inconsistent indentation")
        else:
            end_column = column + len(text)
        if token_type in _OPEN_PARENTHESIS:
            parenthesis_level += 1
        elif token_type in _CLOSE_PARENTHESIS:
            parenthesis_level -= 1
        yield Token(token_type, text, match, row, column, end_row, end_column)
        row = end_row
        column = end_column
    for _ in indents[1:]:
        yield Token(
            TokenType.DEDENT,
            "",
            match,
            end_row,
            end_column,
            end_row,
            end_column,
        )
    yield Token(
        TokenType.END_OF_FILE,
        "",
        match,
        end_row,
        end_column,
        end_row,
        end_column,
    )


if __name__ == "__main__":
    print("|".join(sorted(KEYWORDS)))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import abc
import functools

from ._engine import zones as _zones


ConstantT = t.TypeVar("ConstantT", bound=t.Union[int, float])


@functools.total_ordering
class Bound(t.Generic[ConstantT]):
    _bound: t.Any

    def __init__(self, is_strict: bool, constant: ConstantT) -> None:
        self._bound = _zones.Bound(is_strict, constant)

    def __repr__(self) -> str:
        return f"Bound(is_strict={self.is_strict!r}, constant={self.constant!r})"

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, Bound):
            return self.is_strict == other.is_strict and self.constant == other.constant
        else:
            return False

    def __hash__(self) -> int:
        return hash((Bound, self.is_strict, self.constant))

    def __lt__(self, other: t.Any) -> bool:
        if isinstance(other, Bound):
            if self.constant < other.constant:
                return True
            elif self.constant == other.constant:
                return self.is_strict and not other.is_strict
            else:
                return False
        else:
            return NotImplemented

    @property
    def is_strict(self) -> bool:
        return self._bound.is_strict

    @property
    def constant(self) -> ConstantT:
        return self._bound.constant


def _wrap_bound(bound: t.Any) -> Bound[ConstantT]:
    wrapped = Bound.__new__(Bound)
    wrapped._bound = bound
    return wrapped


class Constraint(t.Generic[ConstantT]):
    _constraint: t.Any

    def __init__(self, left: int, right: int, bound: Bound[ConstantT]) -> None:
        self._constraint = _zones.Constraint(left, right, bound._bound)

    def __repr__(self) -> str:
        return f"Constraint(left={self.left!r}, right={self.right!r}, bound={self.bound!r})"

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, Constraint):
            return (
                self.left == other.left
                and self.right == other.right
                and self.bound == other.bound
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash((Constraint, self.left, self.right, self.bound))

    @property
    def left(self) -> int:
        return self._constraint.left

    @property
    def right(self) -> int:
        return self._constraint.right

    @property
    def bound(self) -> Bound[ConstantT]:
        return _wrap_bound(self._constraint.bound)


def _wrap_constraint(constraint: t.Any) -> Constraint[ConstantT]:
    wrapped = Constraint.__new__(Constraint)
    wrapped._constraint = constraint
    return wrapped


class ReadOnlyZone(t.Protocol[ConstantT]):
    _zone: t.Any

    @property
    def num_variables(self) -> int:
        pass

    @property
    def num_clocks(self) -> int:
        pass

    @property
    def is_empty(self) -> bool:
        pass

    @property
    def constraints(self) -> t.Iterator[Constraint[ConstantT]]:
        pass

    def get_constraint(self, left: int, right: int) -> Constraint[ConstantT]:
        pass

    def get_bound(self, left: int, right: int) -> Bound[ConstantT]:
        pass

    def is_unbounded(self, clock: int) -> bool:
        pass

    def get_upper_bound(self, clock: int) -> t.Optional[ConstantT]:
        pass

    def get_lower_bound(self, clock: int) -> t.Optional[ConstantT]:
        pass

    def is_satisfied(self, constraint: Constraint[ConstantT]) -> bool:
        pass

    def includes(self, other: ReadOnlyZone[ConstantT]) -> bool:
        pass


class Zone(abc.ABC, ReadOnlyZone[ConstantT], t.Generic[ConstantT]):
    _zone: t.Any

    def __init__(self, _zone: t.Any) -> None:
        self._zone = _zone

    @property
    def num_variables(self) -> int:
        return self._zone.num_variables

    @property
    def num_clocks(self) -> int:
        return self._zone.num_clocks

    @property
    def is_empty(self) -> bool:
        return self._zone.is_empty

    @property
    def constraints(self) -> t.Iterator[Constraint[ConstantT]]:
        for left in range(self.num_clocks):
            for right in range(self.num_clocks):
                constraint = self.get_constraint(left, right)
                if constraint.bound.constant is not None:
                    yield constraint

    def get_constraint(self, left: int, right: int) -> Constraint[ConstantT]:
        return _wrap_constraint(self._zone.get_constraint(left, right))

    def get_bound(self, left: int, right: int) -> Bound[ConstantT]:
        return _wrap_bound(self._zone.get_bound(left, right))

    def constrain(self, constraint: Constraint[ConstantT]) -> None:
        self._zone.add_constraint(constraint._constraint)

    def intersect(self, other: Zone[ConstantT]) -> None:
        self._zone.intersect(other._zone)

    def future(self) -> None:
        self._zone.future()

    def past(self) -> None:
        self._zone.past()

    def reset(self, clock: int, value: ConstantT) -> None:
        self._zone.reset(clock, value)

    def is_unbounded(self, clock: int) -> bool:
        return self._zone.is_unbounded(clock)

    def get_upper_bound(self, clock: int) -> t.Optional[ConstantT]:
        return self._zone.get_upper_bound(clock)

    def get_lower_bound(self, clock: int) -> t.Optional[ConstantT]:
        return self._zone.get_lower_bound(clock)

    def is_satisfied(self, constraint: Constraint[ConstantT]) -> bool:
        return self._zone.is_satisfied(constraint._constraint)

    def includes(self, other: ReadOnlyZone[ConstantT]) -> bool:
        return self._zone.includes(other._zone)

    def resize(self, num_variables: int) -> Zone[ConstantT]:
        wrapped = Zone.__new__(Zone)
        wrapped._zone = self._zone.resize(num_variables)
        return wrapped


ZoneT = t.TypeVar("ZoneT", bound=Zone[t.Union[float, int]])


def _wrap_zone(zone: t.Any, cls: t.Type[ZoneT]) -> ZoneT:
    wrapped = cls.__new__(cls)
    wrapped._zone = zone
    return wrapped


class ZoneI64(Zone[int]):
    def __init__(self, num_variables: int) -> None:
        super().__init__(_zones.Zone.new_i64_unconstrained(num_variables))

    @classmethod
    def new_unconstrained(cls, num_variables: int) -> ZoneI64:
        zone = cls.__new__(cls)
        cls._zone = _zones.Zone.new_i64_unconstrained(num_variables)
        return zone

    @classmethod
    def new_zero(cls, num_variables: int) -> ZoneI64:
        zone = cls.__new__(cls)
        cls._zone = _zones.Zone.new_i64_zero(num_variables)
        return zone


class ZoneF64(Zone[float]):
    def __init__(self, num_variables: int) -> None:
        super().__init__(_zones.Zone.new_f64_unconstrained(num_variables))

    @classmethod
    def new_unconstrained(cls, num_variables: int) -> ZoneF64:
        zone = cls.__new__(cls)
        cls._zone = _zones.Zone.new_f64_unconstrained(num_variables)
        return zone

    @classmethod
    def new_zero(cls, num_variables: int) -> ZoneF64:
        zone = cls.__new__(cls)
        cls._zone = _zones.Zone.new_f64_zero(num_variables)
        return zone



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

from . import errors, types
from .actions import ActionPattern
from .automata import Instance
from .expressions import Expression

if t.TYPE_CHECKING:
    from . import context


@d.dataclass(frozen=True, eq=False)
class Link:
    """
    Represents a link of an automaton network.

    Attributes
    ----------
    vector:
        The synchronization vector.
    result:
        The resulting action pattern.
    """

    vector: t.Mapping[Instance, ActionPattern]
    result: t.Optional[ActionPattern] = None
    condition: t.Optional[Expression] = None

    def construct_scope(self, ctx: context.Context) -> context.Scope:
        scope = ctx.global_scope.create_child_scope()
        for pattern in self.vector.values():
            pattern.declare_in(scope)
        return scope


class Network:
    """
    Represents an automaton network.

    Attributes
    ----------
    ctx:
        The :class:`Context` associated with the network.
    name:
        The optional name of the network.
    """

    ctx: context.Context

    name: t.Optional[str]

    _initial_restriction: t.Optional[Expression]

    _instances: t.List[Instance]
    _links: t.List[Link]

    def __init__(self, ctx: context.Context, *, name: t.Optional[str] = None) -> None:
        self.ctx = ctx
        self.name = name
        self._initial_restriction = None
        self._instances = []
        self._links = []

    @property
    def links(self) -> t.List[Link]:
        """
        The set of links of the network.
        """
        return self._links

    @property
    def instances(self) -> t.List[Instance]:
        """
        The set of instances of the network.
        """
        return self._instances

    @property
    def initial_restriction(self) -> t.Optional[Expression]:
        """
        An optional restriction to be satisfied by initial states.

        This property can be set *only once* on a network. The expression
        has to be boolean.

        Raises :class:`~errors.ModelingError` when set twice or to a
        non-boolean expression.
        """
        return self._initial_restriction

    @initial_restriction.setter
    def initial_restriction(self, initial_restriction: Expression) -> None:
        if self._initial_restriction is not None:
            raise errors.ModelingError(
                "restriction of initial valuations has already been set"
            )
        if self.ctx.global_scope.get_type(initial_restriction) != types.BOOL:
            raise errors.InvalidTypeError(
                "restriction of initial valuations must have type `types.BOOL`"
            )
        self._initial_restriction = initial_restriction

    def add_instance(self, instance: Instance) -> None:
        """
        Adds an instance to the network.

        The instance and network are required to be in the same
        modeling context.
        """
        assert instance.automaton.ctx is self.ctx
        if instance not in self._instances:
            self._instances.append(instance)

    def create_link(
        self,
        vector: t.Mapping[Instance, ActionPattern],
        *,
        result: t.Optional[ActionPattern] = None,
        condition: t.Optional[Expression] = None,
    ) -> Link:
        """
        Creates a synchronization link between automata instances.

        The parameter `vector` is a mapping from the instances
        participating in the synchronization to action patterns with
        which they participate.

        The parameter `result` is the action pattern resulting
        from synchronizing.
        """
        link = Link(vector, result=result, condition=condition)
        for instance in vector.keys():
            self.add_instance(instance)
        self._links.append(link)
        return link



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

try:
    import momba_engine as engine

    zones = engine.zones
except ImportError:
    raise ImportError(
        "Missing optional dependency `momba_engine`.\n"
        "Using Momba's engine requires installing `momba_engine`."
    )


__all__ = ["engine", "zones"]
